#pragma once

#include "Drawable.h"
#include "TimeTaker.h"

class Engine
{
public:
	static Engine* Instance();

	// init
	void Initialize(int width, int height, int surfaceSize);
	void Run();
	void Add(IDrawable *drawable);

	// callback functions
	//void onDisplay();
	void onReshape(GLint w, GLint h);
	void onIdle();
	void onKeyboard(unsigned char key, GLint x_coord, GLint y_coord);
	void onSpecial(GLint key, GLint x_coord, GLint y_coord);

	void render();
	unsigned int GetTexCubeMap() { return texCubeMap; }
	float* GetInvModelViewMatrix() { return invmat; }
	float* GetEyeCoord() { return eye_coord; }

private:
	Engine(void);
	~Engine(void);

	void updateCubeMap(unsigned int map);
	void renderSpacedBitmapString(float x, float y, int spacing, void *font, char *string);
	
	void renderScene(bool onlyReflecters);
	void makeInvModelViewMat();

	void printStats(int frames);

	void aviLoop();
	bool writeTGA(const char *file, unsigned char *outImage);

	static Engine* instance;
	std::list<IDrawable*> drawlist;
	std::list<IKeyEvent*> keylist;
	std::list<IStatistics*> statlist;

	int vpWidth;
	int vpHeight;
	int size;

	bool paused;
	bool unpaused;
	bool framebyframe;
	bool firstframe;
	bool fbf_advance;
	int fbf_ticks;
	bool moved;

	float eye_coord[3];
	float line_of_sight[3];
	char fps[10];
	float invmat[16];

	unsigned int texCubeMap;

	bool aviMode;
	unsigned int aviFPS;
	unsigned int aviFrames;

	TimeTaker ttTotal;
};

// GLUT callback functions
void gcDisplay(void);
void gcReshape(GLint w, GLint h);
void gcIdle(void);
void gcKeyboard(unsigned char key, GLint x_coord, GLint y_coord);
void gcSpecial(GLint key, GLint x_coord, GLint y_coord);
