#include "StdAfx.h"
#include "EnvironmentMap.h"

using namespace std;

EnvironmentMap::EnvironmentMap(const string &name, const string &type, int textureSize)
: size(textureSize)
{
	// Create texture and set up states
	glGenTextures(1, &texEnvMap);
	glBindTexture(GL_TEXTURE_CUBE_MAP, texEnvMap);
	
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);

	string prefix = string("media/") + name + string("_");

	loadTexture(GL_TEXTURE_CUBE_MAP_POSITIVE_X, prefix + string("posx.") + type);
	loadTexture(GL_TEXTURE_CUBE_MAP_POSITIVE_Y, prefix + string("posy.") + type);
	loadTexture(GL_TEXTURE_CUBE_MAP_POSITIVE_Z, prefix + string("posz.") + type);
	loadTexture(GL_TEXTURE_CUBE_MAP_NEGATIVE_X, prefix + string("negx.") + type);
	loadTexture(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, prefix + string("negy.") + type);
	loadTexture(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, prefix + string("negz.") + type);

	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP);
	glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP);

	glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
}

void EnvironmentMap::loadTexture(GLenum target, const string &filename)
{
	ILuint ilSky;

	ilGenImages(1, &ilSky); if (!checkILError("GenImages")) return;
	ilBindImage(ilSky); if (!checkILError("BindImage")) return;
	ilLoadImage((ILstring)filename.c_str()); if (!checkILError("LoadImage")) return;
	glTexImage2D(target, 0, GL_RGB, size, size, 0, GL_RGB, GL_UNSIGNED_BYTE, ilGetData());
	ilDeleteImages(1, &ilSky);
}

bool EnvironmentMap::checkILError(const char *location)
{
	ILenum errorNum = ilGetError();
	if (errorNum != IL_NO_ERROR)
	{
		cout << "DevIL error (" << location << "): " << iluErrorString(errorNum) << endl;
	}
	return errorNum == IL_NO_ERROR;
}

void EnvironmentMap::display()
{
	glTranslatef(size / 2.0, 0.0, (size-1.0f)/2.0f);

	float dist = 1000.0f;
	
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);

	glEnable(GL_TEXTURE_CUBE_MAP);
	glBindTexture(GL_TEXTURE_CUBE_MAP, texEnvMap);

	// Front
	glBegin(GL_QUADS);
		glNormal3f(0.0f, 0.0f, -1.0f);
		glTexCoord3f( 1.0f,  1.0f, 1.0f); glVertex3f(-dist, -dist, dist);
		glTexCoord3f( 1.0f, -1.0f, 1.0f); glVertex3f(-dist,  dist, dist);
		glTexCoord3f(-1.0f, -1.0f, 1.0f); glVertex3f( dist,  dist, dist);
		glTexCoord3f(-1.0f,  1.0f, 1.0f); glVertex3f( dist, -dist, dist);
	glEnd();

	// Back
	glBegin(GL_QUADS);
		glNormal3f(0.0f, 0.0f, 1.0f);
		glTexCoord3f( 1.0f,  1.0f, -1.0f); glVertex3f(-dist, -dist, -dist);
		glTexCoord3f(-1.0f,  1.0f, -1.0f); glVertex3f( dist, -dist, -dist);
		glTexCoord3f(-1.0f, -1.0f, -1.0f); glVertex3f( dist,  dist, -dist);
		glTexCoord3f( 1.0f, -1.0f, -1.0f); glVertex3f(-dist,  dist, -dist);
	glEnd();

	// Left
	glBegin(GL_QUADS);
		glNormal3f(1.0f, 0.0f, 0.0f);
		glTexCoord3f(-1.0f,  1.0f, -1.0f); glVertex3f(-dist, -dist,  dist);
		glTexCoord3f(-1.0f,  1.0f,  1.0f); glVertex3f(-dist, -dist, -dist);
		glTexCoord3f(-1.0f, -1.0f,  1.0f); glVertex3f(-dist,  dist, -dist);
		glTexCoord3f(-1.0f, -1.0f, -1.0f); glVertex3f(-dist,  dist,  dist);
	glEnd();

	// Right
	glBegin(GL_QUADS);
		glNormal3f(-1.0f, 0.0f, 0.0f);
		glTexCoord3f(1.0f,  1.0f,  1.0f); glVertex3f(dist, -dist, -dist);
		glTexCoord3f(1.0f,  1.0f, -1.0f); glVertex3f(dist, -dist,  dist);
		glTexCoord3f(1.0f, -1.0f, -1.0f); glVertex3f(dist,  dist,  dist);
		glTexCoord3f(1.0f, -1.0f,  1.0f); glVertex3f(dist,  dist, -dist);
	glEnd();

	// Top
	glBegin(GL_QUADS);
		glNormal3f(0.0f, -1.0f, 0.0f);
		glTexCoord3f( 1.0f, 1.0f, -1.0f); glVertex3f( dist, dist, -dist);
		glTexCoord3f( 1.0f, 1.0f,  1.0f); glVertex3f( dist, dist,  dist);
		glTexCoord3f(-1.0f, 1.0f,  1.0f); glVertex3f(-dist, dist,  dist);
		glTexCoord3f(-1.0f, 1.0f, -1.0f); glVertex3f(-dist, dist, -dist);
	glEnd();

	// Bottom
	glBegin(GL_QUADS);
		glNormal3f(0.0f, 1.0f, 0.0f);
		glTexCoord3f( 1.0f, -1.0f,  1.0f); glVertex3f(-dist, -dist, -dist);
		glTexCoord3f( 1.0f, -1.0f, -1.0f); glVertex3f(-dist, -dist,  dist);
		glTexCoord3f(-1.0f, -1.0f, -1.0f); glVertex3f( dist, -dist,  dist);
		glTexCoord3f(-1.0f, -1.0f,  1.0f); glVertex3f( dist, -dist, -dist);
	glEnd();

	glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
	glDisable(GL_TEXTURE_CUBE_MAP);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
}

void EnvironmentMap::printStatistics()
{
}