#pragma once

#include "Drawable.h"
#include "Shader.h"
#include "TimeTaker.h"

static int perm[512];
static int p[] = {151,160,137,91,90,15,
  131,13,201,95,96,53,194,233,7,225,140,36,103,30,69,142,8,99,37,240,21,10,23,
  190, 6,148,247,120,234,75,0,26,197,62,94,252,219,203,117,35,11,32,57,177,33,
  88,237,149,56,87,174,20,125,136,171,168, 68,175,74,165,71,134,139,48,27,166,
  77,146,158,231,83,111,229,122,60,211,133,230,220,105,92,41,55,46,245,40,244,
  102,143,54, 65,25,63,161, 1,216,80,73,209,76,132,187,208, 89,18,169,200,196,
  135,130,116,188,159,86,164,100,109,198,173,186, 3,64,52,217,226,250,124,123,
  5,202,38,147,118,126,255,82,85,212,207,206,59,227,47,16,58,17,182,189,28,42,
  223,183,170,213,119,248,152, 2,44,154,163, 70,221,153,101,155,167, 43,172,9,
  129,22,39,253, 19,98,108,110,79,113,224,232,178,185, 112,104,218,246,97,228,
  251,34,242,193,238,210,144,12,191,179,162,241, 81,51,145,235,249,14,239,107,
  49,192,214, 31,181,199,106,157,184, 84,204,176,115,121,50,45,127, 4,150,254,
  138,236,205,93,222,114,67,29,24,72,243,141,128,195,78,66,215,61,156,180
};
static int grad3[][3] =	{{1,1,0},{-1,1,0},{1,-1,0},{-1,-1,0},
						{1,0,1},{-1,0,1},{1,0,-1},{-1,0,-1},
						{0,1,1},{0,-1,1},{0,1,-1},{0,-1,-1}};

class WaterSurface : public IDrawable, public IKeyEvent, public IStatistics
{
public:
	WaterSurface(unsigned int surfaceSize, float spacing, float waterLevel);
	virtual ~WaterSurface();

	void setHeight(unsigned int x, unsigned int z, float value);
	void setSpeed(unsigned int x, unsigned int z, float u, float v);
	virtual void update(int ticks);
	virtual void display();
	virtual bool isReflecter() { return false; }
	virtual void onNormalKey(unsigned char key);
	virtual void onSpecialKey(int key);
	virtual void printStatistics();

protected:
	void checkGLErrors(const char *label);
	virtual void computeDeparture();
	virtual void prepareSolver();
	virtual void solver() = 0;
	virtual void postSolver();
	void makeNormals();
	void drawWater();
	void drawNormals();
	void overlayHeightMap(int x, int y);
	void overlaySpeedU(int x, int y);
	void overlaySpeedV(int x, int y);
	void overlayTable2D(int x, int y, float *table, unsigned int tableSize, float r, float g, float b);
	void printHeight();
	void printTable(const char *name, const float *a);
	float simplexNoise(float xin, float yin);

	float position[3];
	float ambient[4];
	float diffuse[4];
	float specular[4];
	float shininess;

	ShaderProgram reflectionShader;

	GLint drawmode;

	bool drawHeightMap;
	bool drawSpeedU;
	bool drawSpeedV;
	
	unsigned int size; // the size of the surface
	unsigned int size_2; // the size squared
	float dS; // delta size
	float dS_1; // 1.0 / delta size
	float dS_2; // delta size squared
	float dS_2_1; // 1.0 / delta size squared
	float dT; // delta t
	float g; // gravitational constant
	float *A;
	float *b;
	float *h;
	float *u;
	float *v;
	float *x;
	float *z;
	float *tildeH;
	float *tildeU;
	float *tildeV;
	float *normals;
	float *noise;

	TimeTaker ttDeparture;
	TimeTaker ttPrepSolver;
	TimeTaker ttSolver;
	TimeTaker ttPostSolver;
	TimeTaker ttDraw;
};

class WaterSurfaceJacobi : public WaterSurface
{
public:
	WaterSurfaceJacobi(unsigned int surfaceSize, float spacing, float waterLevel);

protected:
	virtual void solver();
	void jacobi(int imax, float tolerance);
	float sqrdist(float *x, float *y);
};

class WaterSurfaceLU : public WaterSurface
{
public:
	WaterSurfaceLU(unsigned int surfaceSize, float spacing, float waterLevel);
	~WaterSurfaceLU();
	
protected:
	virtual void solver();
	void LU_solve();
	void LU_decomp();
	void a_to_matrix();

	float *a_mat;
	float *lower;
	float *upper;
};