#include "StdAfx.h"
#include "WaterSurfaceGPU.h"

WaterSurfaceGPU::WaterSurfaceGPU(unsigned int surfaceSize, float spacing, float waterLevel)
: WaterSurface(surfaceSize, spacing, waterLevel), deppointFBO(surfaceSize, surfaceSize),
  ttArray("Create array"), ttTexImage("Tex img"), ttOverhead("GPU overhead"), ttShader("Shaders")
{
	deppointShader.addVertexSource("default.vert");
	deppointShader.addFragmentSource("deppoint.frag");

	deppointValuesTex = createTexture(NULL);
	deppointResultTex = createTexture(NULL);

	rgbaData = new float[size*size*4];
}

WaterSurfaceGPU::~WaterSurfaceGPU()
{
	glDeleteTextures(1, &deppointValuesTex);
	glDeleteTextures(2, &deppointResultTex);
	delete[] rgbaData;
}

void WaterSurfaceGPU::setupView(unsigned int width, unsigned int height)
{
	// setup viewport for 1:1 pixel:texel
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0.0, width, 0.0, height);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glViewport(0, 0, width, height);

	glPolygonMode(GL_FRONT, GL_FILL); // just to be sure
}

unsigned int WaterSurfaceGPU::createTexture(unsigned int width, unsigned int height, const float *data)
{
	// create and bind texture
	unsigned int texture;

	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);

	// turn off filtering and wrap modes
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	// initialize texture data
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, width, height, 0, GL_RGBA, GL_FLOAT, data);

	// unbind texture
	glBindTexture(GL_TEXTURE_2D, 0);
	return texture;
}

unsigned int WaterSurfaceGPU::createTexture(const float *data)
{
	return createTexture(size, size, data);
}

unsigned int WaterSurfaceGPU::createPackedTexture(const float *data)
{
	return createTexture(size/2, size/2, data);
}

void WaterSurfaceGPU::writeTexture(unsigned int texture, 
								   const float *r, const float *g, const float *b, const float *a)
{
	// assumes texture has been created

	// bind texture
	glBindTexture(GL_TEXTURE_2D, texture);

	//float *rgbaData = new float[size_2*4];
	//memset(rgbaData, 0, size_2*4*sizeof(float));

	for (unsigned int i = 0; i < size_2; i++)
	{
		if (r != NULL)
			rgbaData[i*4] = r[i];
		if (g != NULL)
			rgbaData[i*4+1] = g[i];
		if (b != NULL)
			rgbaData[i*4+2] = b[i];
		if (a != NULL)
			rgbaData[i*4+3] = a[i];
	}

	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, size, size, GL_RGBA, GL_FLOAT, rgbaData);
	//delete[] rgbaData;
}

void WaterSurfaceGPU::writePackedTexture(unsigned int texture, const float *data)
{
	// assumes texture has been created

	//float *tdata = new float[size_2]; // size/2 * size/2 * 4 == size*size == size_2
	for (unsigned int i = 0; i < size_2; i++)
	{
		rgbaData[i] = data[i];
	}

	glBindTexture(GL_TEXTURE_2D, texture);
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, size/2, size/2, GL_RGBA, GL_FLOAT, rgbaData);
	//delete[] tdata;
}

void WaterSurfaceGPU::copyTexture(FBO &fbo, unsigned int texOriginal, unsigned int texCopy)
{
	if (!fbo.attach(0, texOriginal))
	{
		return;
	}

	glBindTexture(GL_TEXTURE_2D, texCopy);
	glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
	glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, 0, 0, size, size, 0);
}
/*
void WaterSurfaceGPU::computeDeparture()
{
	//WaterSurface::computeDeparture();

	checkGLErrors("start departure");

	deppointShader.initialize2D(size, size);

	// bind framebuffer object
	deppointFBO.bind();

	// write to read-texture, x = h, y = u, z = v
	writeTexture(deppointValuesTex, h, u, v, NULL);
	checkGLErrors("write texture");

	// attach write-texture to color attachment
	deppointFBO.attach(0, deppointResultTex);
	checkGLErrors("setup fbo");

	// load and run deppoint shader
	deppointShader.load();

	deppointShader.setSampler2D("tex", GL_TEXTURE0, deppointValuesTex);
	deppointShader.setUniform1i("size", size);
	deppointShader.setUniform1f("dT", dT);
	deppointShader.setUniform1f("dS", dS);

	deppointShader.draw2D(0);

	deppointShader.unload();
	checkGLErrors("deppoint shader");

	glBindTexture(GL_TEXTURE_2D, 0);

	// read the result
	deppointFBO.read(0, tildeH, tildeU, tildeV, NULL);
	checkGLErrors("readFBO");

	// restore current viewport
	deppointFBO.unbind();

	deppointShader.reset();

	checkGLErrors("end deppoint");
}
*/
void WaterSurfaceGPU::printStatistics()
{
	WaterSurface::printStatistics();

	std::cout << "Water surface GPU" << std::endl;
	ttOverhead.PrintTime();	ttOverhead.Reset();
	ttShader.PrintTime();	ttShader.Reset();
}