uniform vec3 EyeCoord;
uniform float WaterSize;

const float HalfSize = WaterSize / 2.0;
const float Index_external = 1.0;
const float Index_internal = 1.33;

const float Eta = Index_external / Index_internal; // Ratio of indices of refraction 

const float F  = 0.02040816; // ((1.0-Eta) * (1.0-Eta)) / ((1.0+Eta) * (1.0+Eta));

varying vec3  Reflect;
varying vec3  Refract;
varying float Ratio;
varying bool  Wall;

void main()
{
	vec4 ecPosition  = gl_ModelViewMatrix * gl_Vertex;
	vec3 ecPosition3 = ecPosition.xyz / ecPosition.w;

	vec3 i = normalize(ecPosition3);
	vec3 n = normalize(gl_NormalMatrix * gl_Normal);

	vec3 refr = refract(normalize(gl_Vertex.xyz - EyeCoord), normalize(gl_Normal), Eta);
	vec3 refl = reflect(normalize(gl_Vertex.xyz - EyeCoord), normalize(gl_Normal));

	float dist;
	vec3 refl_point, refr_point;

	if (refl.x < 0.0) {
		dist = -gl_Vertex.x / refl.x;
		refl_point = gl_Vertex.xyz + dist * refl;
	} else {
		dist = (WaterSize - 1.0 - gl_Vertex.x) / refl.x;
		refl_point = gl_Vertex.xyz + dist * refl;
	}
	if (refl_point.y < 0.0) {
		dist = (-HalfSize - gl_Vertex.y) / refl.y;
		refl_point = gl_Vertex.xyz + dist * refl;
	} else if (refl_point.y > WaterSize - 1.0) {
		dist = (HalfSize - gl_Vertex.y) / refl.y;
		refl_point = gl_Vertex.xyz + dist * refl;
	}
	if (refl_point.z < 0.0) {
		dist = -gl_Vertex.z / refl.z;
		refl_point = gl_Vertex.xyz + dist * refl;
	} else if (refl_point.z > WaterSize - 1.0) {
		dist = (WaterSize - 1.0 - gl_Vertex.z) / refl.z;
		refl_point = gl_Vertex.xyz + dist * refl;
	}

	if (refr.x < 0.0) {
		dist = -gl_Vertex.x / refr.x;
		refr_point = gl_Vertex.xyz + dist * refr;
	} else {
		dist = (WaterSize - 1.0 - gl_Vertex.x) / refr.x;
		refr_point = gl_Vertex.xyz + dist * refr;
	}
	if (refr_point.y < 0.0) {
		dist = (-HalfSize - gl_Vertex.y) / refr.y;
		refr_point = gl_Vertex.xyz + dist * refr;
	} else if (refr_point.y > WaterSize - 1.0) {
		dist = (HalfSize - gl_Vertex.y) / refr.y;
		refr_point = gl_Vertex.xyz + dist * refr;
	}
	if (refr_point.z < 0.0) {
		dist = -gl_Vertex.z / refr.z;
		refr_point = gl_Vertex.xyz + dist * refr;
	} else if (refr_point.z > WaterSize - 1.0) {
		dist = (WaterSize - 1.0 - gl_Vertex.z) / refr.z;
		refr_point = gl_Vertex.xyz + dist * refr;
	}

	if (refl_point.y <= 18.0) {
		Reflect = (gl_ModelViewMatrix * (vec4(refl_point ,1) - vec4(HalfSize-0.5, 0.0, HalfSize-0.5, 1.0))).xyz;
		Wall = true;
	} else {
		Reflect = reflect(i, n);
		Wall = false;
	}
	Reflect = vec3(gl_TextureMatrix[0] * vec4(Reflect, 1.0));

	Refract = (gl_ModelViewMatrix * (vec4(refr_point ,1) - vec4(HalfSize, 0.0, HalfSize, 1.0))).xyz;
	Refract = vec3(gl_TextureMatrix[0] * vec4(Refract, 1.0));

	float costhetai = dot(refl, normalize(gl_Normal));
	float costhetat = dot(refr, -normalize(gl_Normal));

	float fresnel_rs = (Index_external * costhetai - Index_external * costhetat)
		/ (Index_external * costhetai + Index_external * costhetat);
	float fresnel_rp = (Index_external * costhetat - Index_external * costhetai)
		/ (Index_external * costhetat + Index_external * costhetai);

	Ratio = min((fresnel_rs * fresnel_rs + fresnel_rp * fresnel_rp) * 0.5, 1.0);

	
	gl_Position = ftransform();
}
