//uniform vec3 EyeCoord;
//uniform float WaterSize;

const float Eta = 0.75;         // Ratio of indices of refraction 
const float FresnelPower = 0.5;

const float F  = ((1.0-Eta) * (1.0-Eta)) / ((1.0+Eta) * (1.0+Eta));

varying vec3  Reflect;
varying vec3  Refract;
varying float Ratio;

void main()
{
	vec3 EyeCoord = vec3(0.0, 40.0, 64.0);
	float WaterSize = 128.0;
    float halfSize = WaterSize / 2.0;
    
    vec4 ecPosition  = gl_ModelViewMatrix * gl_Vertex;
    vec3 ecPosition3 = ecPosition.xyz / ecPosition.w;
    
    vec3 i = normalize(ecPosition3);
    vec3 n = normalize(gl_NormalMatrix * gl_Normal);

	Ratio = F + (1.0 - F) * pow((1.0 - dot(-i, n)), FresnelPower);

    vec3 refr = refract(normalize(gl_Vertex.xyz - EyeCoord), normalize(gl_Normal), Eta);
    vec3 refl = reflect(normalize(gl_Vertex.xyz - EyeCoord), normalize(gl_Normal));
    
    float dist;
    vec3 refl_point, refr_point;
    
    if (gl_Vertex.x + refl.x < 0.0) {
    	dist = (0.0 - gl_Vertex.x) / refl.x;
    	refl_point = gl_Vertex.xyz + dist * refl;
    } else {
    	dist = (WaterSize - 1.0 - gl_Vertex.x) / refl.x;
    	refl_point = gl_Vertex.xyz + dist * refl;
    }
    if (refl_point.y < 0.0) {
    	dist = (-halfSize - gl_Vertex.y) / refl.y;
    	refl_point = gl_Vertex.xyz + dist * refl;
    } else if (refl_point.y > WaterSize - 1.0) {
    	dist = (halfSize - gl_Vertex.y) / refl.y;
    	refl_point = gl_Vertex.xyz + dist * refl;
    }
    if (refl_point.z < 0.0) {
    	dist = (0.0 - gl_Vertex.z) / refl.z;
    	refl_point = gl_Vertex.xyz + dist * refl;
    } else if (refl_point.z > WaterSize - 1.0) {
    	dist = (WaterSize - 1.0 - gl_Vertex.z) / refl.z;
    	refl_point = gl_Vertex.xyz + dist * refl;
    }
    
    if (gl_Vertex.x + refr.x < 0.0) {
    	dist = (0.0 - gl_Vertex.x) / refr.x;
    	refr_point = gl_Vertex.xyz + dist * refr;
    } else {
    	dist = (WaterSize - 1.0 - gl_Vertex.x) / refr.x;
    	refr_point = gl_Vertex.xyz + dist * refr;
    }
    if (refr_point.y < 0.0) {
    	dist = (-halfSize - gl_Vertex.y) / refr.y;
    	refr_point = gl_Vertex.xyz + dist * refr;
    } else if (refr_point.y > WaterSize - 1.0) {
    	dist = (halfSize - gl_Vertex.y) / refr.y;
    	refr_point = gl_Vertex.xyz + dist * refr;
    }
    if (refr_point.z < 0.0) {
    	dist = (0.0 - gl_Vertex.z) / refr.z;
    	refr_point = gl_Vertex.xyz + dist * refr;
    } else if (refr_point.z > WaterSize - 1.0) {
    	dist = (WaterSize - 1.0 - gl_Vertex.z) / refr.z;
    	refr_point = gl_Vertex.xyz + dist * refr;
    }

    
    Reflect = (gl_ModelViewMatrix * (vec4(refl_point ,1) - vec4(halfSize, 0.0, halfSize, 1.0))).xyz;
    Reflect = vec3(gl_TextureMatrix[0] * vec4(Reflect, 1.0));
    
    Refract = (gl_ModelViewMatrix * (vec4(refr_point ,1) - vec4(halfSize, 0.0, halfSize, 1.0))).xyz;
    Refract = vec3(gl_TextureMatrix[0] * vec4(Refract, 1.0));

    gl_Position = ftransform();
}
