#include "lptest2.h"

bool test = true;
bool printTex = false;
bool useShaders = true;
bool cg_on = false;
bool rain = false;
bool printFPS = true;
bool well = false;
bool pause = false;
bool showNormals = false;
GLenum frontmode = GL_LINE;

GLint ind = 0;

static const GLint size = 256;
static const GLint size2 = size * size;
GLfloat delta_size = 0.5f;
GLfloat norm_s_2 = 2.0f * delta_size;
GLfloat delta_size_1 = 1.0f / delta_size;
GLfloat g = 9.8f; // gravitational constant
GLfloat h0 = 6.0f; // homogeneous height at start
GLfloat u0 = 0.0f; // homogeneous start wave velocity in x-direction
GLfloat v0 = 0.0f; // homogeneous start wave velocity in z-direction
GLfloat delta_t; // uses system clock

float h[size2*4];
float h_temp1[size2*4];
float h_temp2[size2*4];

GLuint v_shader, shader_prog, f_shader;
GLuint texID[2], texHandle, sizeHandle, delta_sizeHandle, timeHandle;
GLint currentDrawbuf;

GLuint fboId;
GLint  savedFboId;

GLenum texFormat = GL_RGBA;
GLenum texInternalFormat = GL_RGBA32F_ARB;
GLenum texTarget = GL_TEXTURE_2D;

vector<GLfloat> normals[size][size];
vector<GLfloat> b(size2); // b from Ax = b
matrix<GLfloat> bottom(size, size);
GLfloat bottom_height = 1.0f;
matrix<GLfloat> x(size, size);
matrix<GLfloat> z(size, size);
matrix<GLfloat> A(5, size2);
matrix<GLfloat> tilde_h(size, size);
matrix<GLfloat> tilde_u(size, size);
matrix<GLfloat> tilde_v(size, size);

GLfloat mat_water_diffuse[] = { 0.2f, 0.5f, 1.0f, 0.8f };
GLfloat mat_water_ambient[] = { 0.3f, 0.7f, 1.0f, 1.0f };
GLfloat mat_water_specular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
GLfloat mat_water_shininess[] = { 50.0 };
GLfloat mat_well_diffuse[] = { 0.5f, 0.0f, 0.0f, 1.0f };
GLfloat mat_well_ambient[] = { 0.5f, 0.0f, 0.0f, 1.0f };
GLfloat mat_well_specular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
GLfloat mat_well_shininess[] = { 100.0 };
GLfloat mat_box_diffuse[] = { 0.0f, 0.5f, 0.0f, 1.0f };
GLfloat mat_box_ambient[] = { 0.0f, 0.5f, 0.0f, 1.0f };
GLfloat mat_box_specular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
GLfloat mat_box_shininess[] = { 100.0 };

GLint totalNormal = 0;
GLint totalDraw = 0;
GLint totalDepPoint = 0;
GLint totalPreCG = 0;
GLint totalCG = 0;
GLint totalPostCG = 0;

void init(void)
{
	srand((unsigned) time(NULL));
	GLfloat ambient_light[] = { 0.2f, 0.2f, 0.2f, 1.0f };
	GLfloat light_position[] = { 0.0f, 30.0f, 40.0f, 1.0f };
	
	glClearColor (1.0f, 1.0f, 1.0f, 1.0f);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_POSITION, light_position);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient_light);

	glEnable(GL_DEPTH_TEST);
	glCullFace (GL_BACK);
	glEnable (GL_CULL_FACE);
	glShadeModel (GL_SMOOTH);

	for (GLint n = 0; n < size; n++) {
		h[n*4+1] = 0.0f;
		h[((size-1)*size+n)*4 + 1] = 0.0f;
	}
	for (GLint n = 0; n < size; n++) {
		h[n*size*4 + 2] = 0.0f;
		h[(n*size+(size-1))*4 + 2] = 0.0f;
	}
	
	for (GLint i = 0 ; i < size; i++) {
		for (GLint j = 0; j < size; j++) {
			x(i,j) = i*delta_size;
			z(i,j) = j*delta_size;
			h[(i*size+j)*4] = h0;
			/*printf("[%d][%d][0]=[%d] ", i, j, (i*size+j)*4);
			printf("[%d][%d][1]=[%d] ", i, j, (i*size+j)*4+1);
			printf("[%d][%d][2]=[%d] ", i, j, (i*size+j)*4+2);
			printf("[%d][%d][3]=[%d]\n", i, j, (i*size+j)*4+3);*/
			/*if (i >= size / 64)
				h[i][j][1] = 0.5;
			if (j < size / 64)
				h[i][j][2] = 0.5;*/
			h[(i*size+j)*4+1] = u0;
			h[(i*size+j)*4+2] = v0;
			normals[i][j] = vector<GLfloat>(3);
		}
		//printf("\n");
	}
	//std::cin.get();

	A.clear();
	/*for (GLint i = 0; i < size; i++)
	{
		for (GLint j = 0; j < size; j++)
		{
            bottom(i, j) = 1.0f;
		}
	}*/

	for (GLint i = 2; i < 8; i++) {
		h[(i*size+6)*4] = 2.0f;
		h[(i*size+7)*4] = 2.0f;
		h[(i*size+8)*4] = 2.0f;
		h[(i*size+9)*4] = 2.0f;
		h[(i*size+10)*4] = 2.0f;
		h[(i*size+11)*4] = 2.0f;
		h[(i*size+12)*4] = 2.0f;
		h[(i*size+13)*4] = 2.0f;
		h[(i*size+14)*4] = 2.0f;
		h[(i*size+15)*4] = 2.0f;
	}

	/*for (GLint i = 20; i < 25; i++) {
		h[(i*size+40)*4] = 30.0f;
		h[(i*size+41)*4] = 30.0f;
		h[(i*size+42)*4] = 30.0f;
		h[(i*size+43)*4] = 30.0f;
		h[(i*size+44)*4] = 30.0f;
		h[(i*size+45)*4] = 30.0f;
		h[(i*size+46)*4] = 30.0f;
	}*/
}

void make_normals(void)
{
	GLint startNormal = glutGet(GLUT_ELAPSED_TIME);

	for (GLint i = 1; i < size-1; i++) {
		for (GLint j = 1; j < size-1; j++) {
			normals[i][j](0) = h[((i-1)*size+j)*4] - h[((i+1)*size+j)*4];
			normals[i][j](1) = norm_s_2;
			normals[i][j](2) = h[(i*size+(j+1))*4] - h[(i*size+(j-1))*4];
			normals[i][j] /= norm_2(normals[i][j]);
		}
	}

	for (GLint i = 0; i < size; i++) {
		normals[i][0] = normals[i][1];
		normals[i][size-1] = normals[i][size-2];
	}
	for (GLint i = 0; i < size; i++) {	
		normals[0][i]=normals[1][i];
		normals[size-1][i]=normals[size-2][i];
	}

	totalNormal += glutGet(GLUT_ELAPSED_TIME) - startNormal;
}

void drawNormals(void)
{
	for (int i = 0; i < size; i++) {
		glColor3f(1.0f, 1.0f, 1.0f);
		glBegin(GL_LINES);
		for (int j = 0; j < size; j++) {
			glVertex3f(x(i,j), h[(i*size+j)*4], z(i,j));
			glVertex3f(x(i,j)+normals[i][j][0], h[(i*size+j)*4]+normals[i][j][1], z(i,j)+normals[i][j][2]);
		}
		glEnd();
	}
}

void drawWater(void)
{
	glPolygonMode(GL_FRONT, frontmode);
	if (well) {
		glEnable (GL_BLEND);
		glDepthMask (GL_FALSE);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE);
	}

	for (GLint i = 0; i < size-1; i++) {
		glBegin(GL_TRIANGLE_STRIP);
		glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_water_diffuse);
		glMaterialfv(GL_FRONT, GL_AMBIENT, mat_water_ambient);
		glMaterialfv(GL_FRONT, GL_SPECULAR, mat_water_specular);
		glMaterialfv(GL_FRONT, GL_SHININESS, mat_water_shininess);
		for (GLint j = 0; j < size; j++) {
			//if (useShaders) {
				glNormal3f(normals[i+1][j][0], normals[i+1][j][1], normals[i+1][j][2]);
				glTexCoord2f(x(i+1,j)/(size*delta_size), z(i+1,j)/(size*delta_size));
				//glTexCoord2f(x(size-i-2,size-j-1)/(size*delta_size), z(size-i-2,size-j-1)/(size*delta_size));
				glVertex3f(x(i+1,j), h[(i+1)*size*4+j*4], z(i+1,j));
				
				glNormal3f(normals[i][j][0], normals[i][j][1], normals[i][j][2]);
				glTexCoord2f(x(i,j)/(size*delta_size), z(i,j)/(size*delta_size));
				//glTexCoord2f(x(size-i-1,size-j-1)/(size*delta_size), z(size-i-1,size-j-1)/(size*delta_size));
				glVertex3f(x(i,j), h[(i*size+j)*4], z(i,j));
			/*}
			else {
				glNormal3f(normals[i+1][j][0], normals[i+1][j][1], normals[i+1][j][2]);
				glTexCoord2f(x(i+1,j)/(size*delta_size), z(i+1,j)/(size*delta_size));
				//glTexCoord2f(x(size-i-2,size-j-1)/(size*delta_size), z(size-i-2,size-j-1)/(size*delta_size));
				glVertex3f(x(i+1,j), h[(i+1)*size*4+j*4], z(i+1,j));
				
				glNormal3f(normals[i][j][0], normals[i][j][1], normals[i][j][2]);
				glTexCoord2f(x(i,j)/(size*delta_size), z(i,j)/(size*delta_size));
				//glTexCoord2f(x(size-i-1,size-j-1)/(size*delta_size), z(size-i-1,size-j-1)/(size*delta_size));
				glVertex3f(x(i,j), h[(i*size+j)*4], z(i,j));
			}*/
		}
		glEnd();
	}

	if (well) {
		glDepthMask (GL_TRUE);
		glDisable (GL_BLEND);
	}
}

vector<GLfloat> mat5VecProd(const matrix<GLfloat> &Amat, const vector<GLfloat> &Bvec)
{
	vector<GLfloat> newVec(size2);
	
	for (GLint i = 0; i < size2; i++) {
		newVec(i)  = column(Amat,i)(0) * Bvec(i);
		if (i < size2-1)
			newVec(i) += column(Amat,i)(1) * Bvec(i+1);
		if (i < size2-size)
			newVec(i) += column(Amat,i)(2) * Bvec(i+size);
		if (i > size)
			newVec(i) += column(Amat,i)(3) * Bvec(i-size);
		if (i > 0)
			newVec(i) += column(Amat,i)(4) * Bvec(i-1);
	}

	return newVec;
}

GLfloat dista(const GLfloat x[size2*4], const GLfloat y[size2*4]) {
	GLfloat sum = 0.0;

	for (GLint i = 0; i < size; i++) {
		for (GLint j = 0; j < size; j++) {
			sum = sum + (x[(i*size+j)*4] - y[i*size*4+j])*(x[i*size*4+j] - y[i*size*4+j]);
		}
	}
	return sqrt(sum);
}

void Jacobi(GLint imax, GLfloat tolerance) {
	GLint iter_num = 0;
	GLfloat sum = 0.0;
	GLfloat x_old[size2*4];

	for (GLint i = 0; i < size; i++) {
		for (GLint j = 0; j < size; j++) {
			h[(i*size+j)*4] = b(i*size+j);
		}
	}

	do {
		iter_num++;
		for (GLint i = 0; i < size; i++) {
			for (GLint j = 0; j < size; j++) {
				x_old[i*size*4+j] = h[(i*size+j)*4];
			}
		}

		for (GLint i = 0; i < size; i++) {
			for (GLint j = 0; j < size; j++) {
				h[(i*size+j)*4] = b(i*size + j);
				if (i*size + j > 0)
					h[(i*size+j)*4] -= A(1,i*size+j-1)*x_old[i*size*4+j];
				if (i*size + j > size)
					h[(i*size+j)*4] -= A(2,(i-1)*size+j)*x_old[i*size*4+j];
				if (i*size + j < size2-size)
					h[(i*size+j)*4] -= A(3,(i+1)*size+j)*x_old[i*size*4+j];
				if (i*size + j < size2-1)
					h[(i*size+j)*4] -= A(4,i*size+j+1)*x_old[i*size*4+j];
				h[(i*size+j)*4] /= A(0,i*size+j);
			}
		}
	} while ((iter_num < imax) && (dista(h,x_old) >= tolerance));
}

/*
 -- Initialize --
 Choose x0
 Set
 i = 0 and imax = maximum number of iteration to be performed
 r^0 = b - Ax^0
 p^0 = r^0

 -- Begin CG iteration --
 If i < imax, Perform Steps 3 - 5
 If i = imax, Exit and print message "i = imax"

 -- Perform CG update --
 Set
 alpha_i = dot(r^i,p^i) / dot(p^i,A*p^i)
 x^(i+1) = x^i + alpha_i * p^i
 r^(i+1) = r^i - alpha_i * A * p^i

 -- check for convergence --
 if ||r^(i+1)|| < tolerance go to step 6

 -- Prepare for next CG update --
 Set
 x^i = x^(i+1)
 r^i = r^(i+1)
 Beta_i = - dot(r^i,A*r^i) / dot(p^i,A*p^i)
 p^i = r^i + Beta_i * p^i
 i = i+1
 Go to step 3
*/
//vector<GLfloat> CG(const matrix<GLfloat> &Amat, const vector<GLfloat> &Bvec, GLint imax, GLfloat tolerance)
//void CG(const matrix<GLfloat> &Amat, const vector<GLfloat> &Bvec, float[size][size] &x, GLint imax, GLfloat tolerance)
//{
//	GLfloat alpha_i;
//	GLfloat beta_i;
//	//vector<GLfloat> x_i(size2);
//	vector<GLfloat> r_i(size2);
//	vector<GLfloat> p_i(size2);
//	vector<GLfloat> a_p_i(size2);
//	GLfloat d_a_p_i;
//
//	r_i = Bvec - mat5VecProd(Amat, x);//prod(Amat, x_i);
//	p_i = r_i;
//	for (GLint i = 0; i < imax; i++)
//	{
//		a_p_i = mat5VecProd(Amat, p_i);//prod(Amat, p_i); // disse endringene gjr at imax kan kes og tolerance senkes uten dramatisk innvirkning p FPS
//		d_a_p_i = inner_prod(p_i, a_p_i);
//
//		alpha_i = inner_prod(r_i,p_i) / d_a_p_i; //inner_prod(p_i, prod(Amat, p_i));
//		x += alpha_i * p_i;
//		if (i == imax-1) {
//			//std::cout << "i == imax-1 == " << i << std::endl;
//			break; /* denne endringen kte FPS fra 38 til 56 */
//		}
//
//		noalias(r_i) -= alpha_i * a_p_i; //prod(Amat, p_i);
//		if (norm_2(r_i) < tolerance) {
//			//std::cout << "tolerance i==" << i << std::endl;
//			break;	/* denne endringen kte FPS fra 28 til 38 */
//		}
//
//		beta_i = - inner_prod(r_i,mat5VecProd(Amat,r_i)/*prod(Amat,r_i)*/) / d_a_p_i; //inner_prod(p_i,prod(Amat,p_i));
//		p_i = r_i + beta_i * p_i;
//	}
//
//	//return x_i;
//}

void wave_algorithm(GLint elapsedTime)
{
	delta_t = 0.005; //elapsedTime / 1000.0f;

	GLint startDepPoint = glutGet(GLUT_ELAPSED_TIME);
	for (GLint i = 0 ; i < size; i++) {
		for (GLint j = 0; j < size; j++) {
			// departure point from (31). Vector since 2D
			GLfloat alpha_x = delta_t * h[(i*size+j)*4 + 1];
			GLfloat alpha_z = delta_t * h[(i*size+j)*4 + 2];

			/* --- denne endringen ga en liten speedup p deppoint-beregningene */
			GLfloat dep_p_x = (x(i,j)-alpha_x) * delta_size_1;
			GLfloat dep_p_z = (z(i,j)-alpha_z) * delta_size_1;

			GLint xm_1, xm, zn_1, zn;
			if (test) {
				tilde_h(i,j) = h_temp2[(i*size + j)*4];
				tilde_u(i,j) = h_temp2[(i*size + j)*4 + 1];
				tilde_v(i,j) = h_temp2[(i*size + j)*4 + 2];
			} else {
				xm_1 = (int)(dep_p_x+0.0000001);
				xm   = xm_1 + 1;
				zn_1 = (int)(dep_p_z+0.0000001);
				zn   = zn_1 + 1;
				if (xm_1 < 0) {
					xm_1 = 0;
					xm = 0;
				} else if (xm < 0) {
					xm_1 = 0;
				} else if (xm >= size) {
					xm = size-1;
					xm_1 = size-1;
				} else if (xm_1 >= size) {
					xm_1 = size-1;
				}
				if (zn_1 < 0) {
					zn_1 = 0;
					zn = 0;
				} else if (zn < 0) {
					zn = 0;
				} else if (zn >= size) {
					zn = size-1;
					zn_1 = size-1;
				} else if (zn_1 >= size) {
					zn_1 = size-1;
				}
			/*	xm_1 = max(xm_1, 0); xm_1 = min(xm_1, size-1);
				xm   = max(xm, 0);	 xm   = min(xm, size-1);
				zn_1 = max(zn_1, 0); zn_1 = min(zn_1, size-1);
				zn   = max(zn, 0);	 zn   = min(zn, size-1);*/
			
				if (xm == xm_1) {
					if (zn == zn_1) {
						tilde_h(i,j) = h[(xm*size+zn)*4];
						tilde_u(i,j) = h[(xm*size+zn)*4+1];
						tilde_v(i,j) = h[(xm*size+zn)*4+2];
					} else {
						GLfloat z1 = (z(i,zn)-z(i,j)+alpha_z)*delta_size_1;
						GLfloat z2 = 1.0f-z1; //(z(i,j)-alpha_z-z(i,zn_1))*delta_size_1;
						tilde_h(i,j) = z1*h[xm*size*4+zn_1*4]   + z2*h[xm*size*4+zn*4];
						tilde_u(i,j) = z1*h[xm*size*4+zn_1*4+1] + z2*h[xm*size*4+zn*4+1];
						tilde_v(i,j) = z1*h[xm*size*4+zn_1*4+2] + z2*h[xm*size*4+zn*4+2];
					}
				} else {
					GLfloat x1 = (x(xm,j)-x(i,j)+alpha_x)*delta_size_1;
					GLfloat x2 = 1.0f-x1; //(x(i,j)-alpha_x-x(xm_1,j))*delta_size_1;
					if (zn == zn_1) {
						tilde_h(i,j) = x1*h[xm_1*size*4+zn*4]   + x2*h[xm*size*4+zn*4];
						tilde_u(i,j) = x1*h[xm_1*size*4+zn*4+1] + x2*h[xm*size*4+zn*4+1];
						tilde_v(i,j) = x1*h[xm_1*size*4+zn*4+2] + x2*h[xm*size*4+zn*4+2];
					} else {
						GLfloat z1 = (z(i,zn)-z(i,j)+alpha_z)*delta_size_1;
						GLfloat z2 = 1.0f-z1; //(z(i,j)-alpha_z-z(i,zn_1))*delta_size_1;
						tilde_h(i,j) = x2*(z1*h[xm*size*4+zn_1*4]   + z2*h[xm*size*4+zn*4])   + x1*(z1*h[xm_1*size*4+zn_1*4]   + z2*h[xm_1*size*4+zn*4]);
						tilde_u(i,j) = x2*(z1*h[xm*size*4+zn_1*4+1] + z2*h[xm*size*4+zn*4+1]) + x1*(z1*h[xm_1*size*4+zn_1*4+1] + z2*h[xm_1*size*4+zn*4+1]);
						tilde_v(i,j) = x2*(z1*h[xm*size*4+zn_1*4+2] + z2*h[xm*size*4+zn*4+2]) + x1*(z1*h[xm_1*size*4+zn_1*4+2] + z2*h[xm_1*size*4+zn*4+2]);
					}
				}

				h_temp1[(i*size + j)*4]     = (1.0f-((x(xm,j)-x(i,j)+alpha_x)*delta_size_1))*(((z(i,zn)-z(i,j)+alpha_z)*delta_size_1)*h[xm*size*4+zn_1*4] + (1.0f-((z(i,zn)-z(i,j)+alpha_z)*delta_size_1))*h[xm*size*4+zn*4])
					+ ((x(xm,j)-x(i,j)+alpha_x)*delta_size_1)*(((z(i,zn)-z(i,j)+alpha_z)*delta_size_1)*h[xm_1*size*4+zn_1*4]   + (1.0f-((z(i,zn)-z(i,j)+alpha_z)*delta_size_1))*h[xm_1*size*4+zn*4]);
				h_temp1[(i*size + j)*4 + 1] = 1.0f - (x(xm,j)-x(i,j)+alpha_x)*delta_size_1;
				h_temp1[(i*size + j)*4 + 2] = zn_1;
				h_temp1[(i*size + j)*4 + 3] = zn;
			}

			if (printTex) {
				if ((h_temp2[(i*size + j)*4 + ind] - h_temp1[(i*size + j)*4 + ind]) > 0.01f) {
					printf("Feil 1: %2d, %2d   %10.7f != %10.7f\n", i, j, h_temp1[(i*size + j)*4 + ind], h_temp2[(i*size + j)*4 + ind]);
				}
			}
		}
	}

	/*for (GLint i = 0 ; i < size; i++) {
		for (GLint j = 0; j < size; j++) {
			h[(i*size + j)*4 + 1] = h_temp2[(i*size + j)*4];
			h[(i*size + j)*4 + 2] = h_temp2[(i*size + j)*4 + 1];
		}
	}*/

	if (printTex) {
		printf("       ");
		for (int i = 0; i < size; i++) {
			printf("%7d", i);
		}
		printf("%\n");
		for (int i = 0; i < size; i++) {
			printf("%2d CPU: ", i);
			for (int j = 0; j < size; j++) {
				printf("%6.2f ", h_temp1[(i*size + j)*4 + ind]);
			}
			printf("\n%2d GPU: ", i);
			for (int j = 0; j < size; j++) {
				printf("%6.2f ", h_temp2[(i*size + j)*4 + ind]);
			}
			printf("\n\n");
		}
		printf("slutt");
		std::cin.get();
	}

	totalDepPoint += glutGet(GLUT_ELAPSED_TIME) - startDepPoint;

	GLint startPreCG = glutGet(GLUT_ELAPSED_TIME);

	GLfloat dij;
	GLfloat x2 = delta_size*delta_size;
	GLfloat t2g = delta_t*delta_t*g;

	/* -- Complete (25) implemented underneath. --
	   Simplification of (25). Should implement complete formula.
		This simplification assumes flat bottom with delta_size == delta_size == 1.0.
		The extra terms on the second line of the formula for A are there
		because the because the waterheight should not change across the
		boundaries. By these extra terms we change the height of the points
		outside the grid to the same height as	the point nearest them inside the grid.
		*/
	/*for (GLint i = 0; i < size; i++) {
		for (GLint j = 0; j < size; j++) {
			dij = h(i*size*4+j)-bottom(i,j);
			b[i*size+j] = tilde_h(i,j)/(t2g*dij)
				- (tilde_u((i==size-1?i:i+1),j) - tilde_u((i==0?0:i-1),j) + tilde_v(i,(j==size-1?j:j+1)) - tilde_v(i,(j==0?0:j-1))) / (2*delta_t*g);
			A(0,i*size+j) = 1.0/(t2g*dij) + 4.0 - (i==0?1.0:0.0) - (j==0?1.0:0.0) - (i==size-1?1.0:0.0) - (j==size-1?1.0:0.0);

			if (i > 0)		A(3,i*size+j) = -1.0;
			if (i < size-1)	A(2,i*size+j) = -1.0;
			if (j > 0)		A(4,i*size+j)	= -1.0;
			if (j < size-1)	A(1,i*size+j)	= -1.0;
		}
	}*/

	/* Another simplification of (25). Assumes flat bottom */
	//for (GLint i = 0; i < size; i++) {
	//	for (GLint j = 0; j < size; j++) {
	//		dij = h(i*size*4+j)-bottom(i,j);
	//		b[i*size+j] = (tilde_h(i,j)*x2)/(t2g*dij)
	//			- delta_size*(tilde_u((i==size-1?i:i+1),j) - tilde_u((i==0?0:i-1),j) + tilde_v(i,(j==size-1?j:j+1)) - tilde_v(i,(j==0?0:j-1))) / (2*delta_t*g);
	//		A(0,i*size+j) = x2/(t2g*dij) /*+ 4.0*/ + (i==0?0.0:1.0) + (j==0?0.0:1.0) + (i==size-1?0.0:1.0) + (j==size-1?0.0:1.0);

	//		if (i > 0)		A(3,i*size+j) = -1.0;
	//		if (i < size-1)	A(2,i*size+j) = -1.0;
	//		if (j > 0)		A(4,i*size+j)	= -1.0;
	//		if (j < size-1)	A(1,i*size+j)	= -1.0;
	//	}
	//}

	GLfloat bottom_diff_x = 0.0f;
	GLfloat bottom_diff_z = 0.0f;

	for (GLint i = 0; i < size; i++) {
		for (GLint j = 0; j < size; j++) {
			dij = h[i*size*4 + j*4]-bottom_height;
			GLfloat temp1 = -(t2g/x2)*dij;
	
			if (i > 0)		A(3,i*size+j) = temp1;
			if (j > 0)		A(4,i*size+j) = temp1;

			A(0,i*size+j) = 1.0f + (4.0f*t2g*dij)/x2
				+ (i==0    ? temp1 : 0.0)
				+ (i==size-1 ? temp1 : 0.0)
				+ (j==0    ? temp1 : 0.0)
				+ (j==size-1 ? temp1 : 0.0);

			if (i < size-1)	A(2,i*size+j) = temp1;
			if (j < size-1)	A(1,i*size+j) = temp1;

			b[i*size+j] = tilde_h(i,j)
				- 0.5f*delta_t*dij*((tilde_u((i==size-1?i:i+1),j)-tilde_u((i==0?0:i-1),j))*delta_size_1
				+ (tilde_v(i,(j==size-1?j:j+1))-tilde_v(i,(j==0?0:j-1)))*delta_size_1);
		}
	}

	// Complete (25) implemented
	/*GLfloat z2 = delta_size*delta_size;
	for (GLint i = 0; i < size; i++) {
		for (GLint j = 0; j < size; j++) {
			dij = h(i*size+j)-bottom(i,j);
			if (dij < 0.0f)
				dij = 0.0f;
			
			GLfloat bottom_diff_x;
			GLfloat bottom_diff_z;

			if (i == 0 || i == size - 1) {
				bottom_diff_x = 0.0f;
			} else {
				bottom_diff_x = bottom(i+1,j)-bottom(i-1,j);
			}
			if (j == 0 || j == size - 1){
				bottom_diff_z = 0.0f;
			} else {
				bottom_diff_z = bottom(i,j+1)-bottom(i,j-1);

			if (i > 0)		A(3,i*size+j) = -(t2g/x2)*(.25f*bottom_diff_x + dij);
			if (j > 0)		A(4,i*size+j)	= -(t2g/z2)*(.25f*bottom_diff_z + dij);

			A(0,i*size+j) = 1.0f + (2.0f*t2g*dij)*(1.0f/x2 + 1.0f/z2)
				+ (i==0    ? -(t2g/x2)*(.25f*bottom_diff_x + dij) : 0.0)
				+ (i==size-1 ?  (t2g/x2)*(.25f*bottom_diff_x - dij) : 0.0)
				+ (j==0    ? -(t2g/z2)*(.25f*bottom_diff_z + dij) : 0.0)
				+ (j==size-1 ?  (t2g/z2)*(.25f*bottom_diff_z - dij) : 0.0);

			if (i < size-1)	A(2,i*size+j) = (t2g/x2)*(.25f*bottom_diff_x - dij);
			if (j < size-1)	A(1,i*size+j) = (t2g/z2)*(.25f*bottom_diff_z - dij);

			b[i*size+j] = tilde_h(i,j)
				+ 0.5f*delta_t*((tilde_u(i,j)*bottom_diff_x)*delta_size_1
					+ (tilde_v(i,j)*bottom_diff_z)*delta_size_1)
				- 0.5f*delta_t*dij*((tilde_u((i==size-1?i:i+1),j)-tilde_u((i==0?0:i-1),j))*delta_size_1
				+ (tilde_v(i,(j==size-1?j:j+1))-tilde_v(i,(j==0?0:j-1)))*delta_size_1);
		}
	}*/
	totalPreCG += glutGet(GLUT_ELAPSED_TIME) - startPreCG;

	// find solution with conjugate gradient method.
	GLint startCG = glutGet(GLUT_ELAPSED_TIME);

	//if (cg_on) {
	//	CG(A, b, h, 1, 2.0f);
	//} else {
		Jacobi(2, 0.001f);
	//}

	totalCG += glutGet(GLUT_ELAPSED_TIME) - startCG;

	// updates h, u and v.
	// need to change the formulas of u and v to (21) and (22)
	GLint startPostCG = glutGet(GLUT_ELAPSED_TIME);
	for (GLint i = 0; i < size; i++) {
		for (GLint j = 0; j < size; j++) {
			if (h[i*size*4 + j*4] < 0.0f)
				h[i*size*4 + j*4] = 0.0f;
		}
	}

	for (GLint i = 1; i < size-1; i++) {
		for (GLint j = 1; j < size-1; j++) {
			h[i*size*4 + j*4+1] = -0.5f*delta_t*g*(h[((i+1)*size + j)*4] - h[((i-1)*size + j)*4])*delta_size_1 + tilde_u(i,j);
			h[i*size*4 + j*4+2] = -0.5f*delta_t*g*(h[(i*size + (j+1))*4] - h[(i*size + (j-1))*4])*delta_size_1 + tilde_v(i,j);
		}
	}
	totalPostCG += glutGet(GLUT_ELAPSED_TIME) - startPostCG;
}

void printStats()
{
	std::cout << "Total normal:    " << totalNormal << std::endl;
	std::cout << "Total draw:      " << totalDraw << std::endl;
	std::cout << "Total dep point: " << totalDepPoint << std::endl;
	std::cout << "Total pre CG:    " << totalPreCG << std::endl;
	std::cout << "Total CG:        " << totalCG << std::endl;
	std::cout << "Total post CG:   " << totalPostCG << std::endl;
}

void idle(void)
{
	//static bool stats_10sec_printed = false;
	//static GLint totalTime = 0;
	//GLint startTime = glutGet(GLUT_ELAPSED_TIME);

	//if (!stats_10sec_printed && totalTime > 10000) {
	//	printStats();
	//	stats_10sec_printed = true;
	//}

	//static GLint oldTime = glutGet(GLUT_ELAPSED_TIME);
	//static GLint timebase = oldTime;
	//static GLint frame = 0;

	//GLint newTime = glutGet(GLUT_ELAPSED_TIME);
	//frame++;

	////std::cout << "Elapsed time: " << newTime - oldTime << std::endl;

	//if (oldTime == newTime)
	//	return;

	//if (newTime - timebase > 1000) {
	//	if (printFPS)
	//		std::cout << "FPS: " << (GLint)(frame*1000.0f/(newTime-timebase)) << std::endl;
	//	timebase = newTime;
	//	frame = 0;
	//}

	//if (!pause) {
	//	// rain
	//	if (rain) {
	//		GLint chance = (GLint) (((GLfloat) rand() / (GLfloat) RAND_MAX) * 10.0f);
	//		if (chance > 6) {
	//			GLint i = (GLint) (((GLfloat) rand() / (GLfloat) RAND_MAX) * (GLfloat) (size-2)) + 1;
	//			GLint j = (GLint) (((GLfloat) rand() / (GLfloat) RAND_MAX) * (GLfloat) (size-2)) + 1;
	//			//std::cout << "i: " << i << "  j: " << j << std::endl;
	//			if (i > 0) h[i-1][j][0] += .5f; else h[i][j][0] += .5f;
	//			if (j > 0) h[i][j-1][0] += .5f; else h[i][j][0] += .5f;
	//			if (i > 0 && j > 0) h[i-1][j-1][0] += .5f; else h[i][j][0] += .5f;
	//			if (j > 0 && i < size-2) h[i+1][j-1][0] += .5f; else h[i][j][0] += .5f;
	//			h[i][j][0] -= 4.0f;
	//			if (i < size-2) h[i+1][j][0] += .5f; else h[i][j][0] += 1.0f;
	//			if (j < size-2) h[i][j+1][0] += .5f; else h[i][j][0] += 1.0f;
	//			if (i < size-2 && j < size-2) h[i+1][j+1][0] += .5f; else h[i][j][0] += .5f;
	//			if (j < size-2 && i > 0) h[i-1][j+1][0] += .5f; else h[i][j][0] += .5f;
	//		}
	//	}

	//	wave_algorithm(newTime - oldTime);
	//}
	//oldTime = newTime;

	glutPostRedisplay();

	//totalTime += glutGet(GLUT_ELAPSED_TIME) - startTime;
}

void drawWell(void)
{
	GLfloat hOverWater = 3.0f;
	glPolygonMode(GL_FRONT, frontmode);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_well_diffuse);
	glMaterialfv(GL_FRONT, GL_AMBIENT, mat_well_ambient);
	glBegin(GL_QUADS);
		glNormal3f(-1.0f, 0.0f, 0.0f);
		glVertex3f(-3.0f,				0.0f,				-3.0f);
		glVertex3f(-3.0f,				0.0f,				(size-1)*delta_size+3.0f);
		glVertex3f(-3.0f,				h0+hOverWater,			(size-1)*delta_size+3.0f);
		glVertex3f(-3.0f,				h0+hOverWater,			-3.0f);
	glEnd();
	glBegin(GL_QUADS);
		glNormal3f(0.0f, 1.0f, 0.0f);
		glVertex3f(-3.0f,				h0+hOverWater,			0.0f);
		glVertex3f(-3.0f,				h0+hOverWater,			(size-1)*delta_size);
		glVertex3f(0.0f,				h0+hOverWater,			(size-1)*delta_size);
		glVertex3f(0.0f,				h0+hOverWater,			0.0f);
	glEnd();
	glBegin(GL_QUADS);
		glNormal3f(1.0f, 0.0f, 0.0f);
		glVertex3f(0.0f,				h0+hOverWater,			0.0f);
		glVertex3f(0.0f,				h0+hOverWater,			(size-1)*delta_size);
		glVertex3f(0.0f,				bottom(0,size-1),		(size-1)*delta_size);
		glVertex3f(0.0f,				bottom(0,0),		0.0f);
	glEnd();
	glBegin(GL_QUADS);
		glNormal3f(0.0f, 1.0f, 0.0f);
		glVertex3f(0.0f,				bottom(0,0),		0.0f);
		glVertex3f(0.0f,				bottom(0,size-1),		(size-1)*delta_size);
		glVertex3f((size-1)*delta_size,		bottom(size-1,size-1),	(size-1)*delta_size);
		glVertex3f((size-1)*delta_size,		bottom(size-1,0),		0.0f);
	glEnd();
	glBegin(GL_QUADS);
		glNormal3f(-1.0f, 0.0f, 0.0f);
		glVertex3f((size-1)*delta_size,		bottom(size-1,size-1),	(size-1)*delta_size);
		glVertex3f((size-1)*delta_size,		h0+hOverWater,			(size-1)*delta_size);
		glVertex3f((size-1)*delta_size,		h0+hOverWater,			0.0f);
		glVertex3f((size-1)*delta_size,		bottom(size-1,0),		0.0f);
	glEnd();
	glBegin(GL_QUADS);
		glNormal3f(0.0f, 1.0f, 0.0f);
		glVertex3f((size-1)*delta_size,		h0+hOverWater,			0.0f);
		glVertex3f((size-1)*delta_size,		h0+hOverWater,			(size-1)*delta_size);
		glVertex3f((size-1)*delta_size+3.0f,	h0+hOverWater,			(size-1)*delta_size);
		glVertex3f((size-1)*delta_size+3.0f,	h0+hOverWater,			0.0f);
	glEnd();
	glBegin(GL_QUADS);
		glNormal3f(1.0f, 0.0f, 0.0f);
		glVertex3f((size-1)*delta_size+3.0f,	h0+hOverWater,			-3.0f);
		glVertex3f((size-1)*delta_size+3.0f,	h0+hOverWater,			(size-1)*delta_size+3.0f);
		glVertex3f((size-1)*delta_size+3.0f,	0.0f,				(size-1)*delta_size+3.0f);
		glVertex3f((size-1)*delta_size+3.0f,	0.0f,				-3.0f);
	glEnd();

	glBegin(GL_QUADS);
		glNormal3f(0.0f, 0.0f, -1.0f);
		glVertex3f(-3.0f,				0.0f,				-3.0f);
		glVertex3f(-3.0f,				h0+hOverWater,			-3.0f);
		glVertex3f((size-1)*delta_size+3.0f,	h0+hOverWater,			-3.0f);
		glVertex3f((size-1)*delta_size+3.0f,	0.0f,				-3.0f);
	glEnd();
	glBegin(GL_QUADS);
		glNormal3f(0.0f, 1.0f, 0.0f);
		glVertex3f(-3.0f,				h0+hOverWater,			-3.0f);
		glVertex3f(-3.0f,				h0+hOverWater,			0.0f);
		glVertex3f((size-1)*delta_size+3.0f,	h0+hOverWater,			0.0f);
		glVertex3f((size-1)*delta_size+3.0f,	h0+hOverWater,			-3.0f);
	glEnd();
	glBegin(GL_QUADS);
		glNormal3f(1.0f, 0.0f, 0.0f);
		glVertex3f(0.0f,				bottom(0,0),		0.0f);
		glVertex3f((size-1)*delta_size,		bottom(size-1,0),		0.0f);
		glVertex3f((size-1)*delta_size,		h0+hOverWater,			0.0f);
		glVertex3f(0.0f,				h0+hOverWater,			0.0f);
	glEnd();
	glBegin(GL_QUADS);
		glNormal3f(0.0f, 0.0f, 1.0f);
		glVertex3f(-3.0f,				0.0f,				(size-1)*delta_size+3.0f);
		glVertex3f((size-1)*delta_size+3.0f,	0.0f,				(size-1)*delta_size+3.0f);
		glVertex3f((size-1)*delta_size+3.0f,	h0+hOverWater,			(size-1)*delta_size+3.0f);
		glVertex3f(-3.0f,				h0+hOverWater,			(size-1)*delta_size+3.0f);
	glEnd();
	glBegin(GL_QUADS);
		glNormal3f(0.0f, 1.0f, 0.0f);
		glVertex3f(-3.0f,				h0+hOverWater,			(size-1)*delta_size+3.0f);
		glVertex3f((size-1)*delta_size+3.0f,	h0+hOverWater,			(size-1)*delta_size+3.0f);
		glVertex3f((size-1)*delta_size+3.0f,	h0+hOverWater,			(size-1)*delta_size);
		glVertex3f(-3.0f,				h0+hOverWater,			(size-1)*delta_size);
	glEnd();
	glBegin(GL_QUADS);
		glNormal3f(-1.0f, 0.0f, 0.0f);
		glVertex3f(0.0f,				bottom(0,size-1),		(size-1)*delta_size);
		glVertex3f(0.0f,				h0+hOverWater,			(size-1)*delta_size);
		glVertex3f((size-1)*delta_size,		h0+hOverWater,			(size-1)*delta_size);
		glVertex3f((size-1)*delta_size,		bottom(size-1,size-1),	(size-1)*delta_size);
	glEnd();
}

void checkGLErrors (const char *label = NULL) {
	GLuint errnum;
	const char *errstr;
	while (errnum = glGetError()) 
	{
		errstr = reinterpret_cast<const char *>(gluErrorString(errnum));
		std::cerr << errstr; 

		if(label) std::cerr << " at " << label;		
		std::cerr << endl;
	}
	return;
}

void display(void)
{
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glColor3f (1.0f, 1.0f, 1.0f);
	glLoadIdentity (); // clear the matrix
	/* viewing transformation */
	glFrustum (-1.0f, 1.0f, -1.0f, 1.0f, 1.5f, 500.0f);
	gluLookAt (40.0f, 40.0f, 50.0f, 25.0f, 20.0f, 25.0f, 0.0f, 1.0f, 0.0f);

	make_normals();

	GLint startDraw = glutGet(GLUT_ELAPSED_TIME);
	
	if (well)
		drawWell();

	//if (useShaders) {
	//	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	//	glDrawBuffer(currentDrawbuf);

	//	glBindTexture(texTarget, texID[1]);
	//	glEnable(GL_TEXTURE_2D);
	//	//glUseProgram(shader_prog);
	//	//glTexImage2D(texTarget, 0, texInternalFormat, size, size, 0, texFormat, GL_FLOAT, h);
	//}

	drawWater();
	
	/*if (useShaders) {
		glDisable(GL_TEXTURE_2D);
		glUseProgram(0);
		glBindTexture(texTarget, 0);
	}*/
	
	if (showNormals)
		drawNormals();

	
	if (useShaders) {
		// setup viewport for 1:1 pixel:texel
		glMatrixMode(GL_PROJECTION);
		glPushMatrix(); // save current viewport state
		
		glLoadIdentity();
		gluOrtho2D(0.0, (GLfloat)size, 0.0, (GLfloat)size);
4		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		glViewport(0, 0, size, size);
		glDisable(GL_CULL_FACE);
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_LIGHTING);

		checkGLErrors ("START: display");
		glGetIntegerv(GL_DRAW_BUFFER, &currentDrawbuf);		// Save the current Draw buffer
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fboId);	// Render to the FBO
		glDrawBuffer(GL_COLOR_ATTACHMENT1_EXT);				// Draw into the second texture

		glClear(GL_COLOR_BUFFER_BIT);
		glMatrixMode(GL_MODELVIEW);
		glPolygonMode(GL_FRONT, GL_FILL);

		//glutSolidTeapot(1.3);

		//checkGLErrors ("display 1");
		
		//glDrawBuffer(GL_COLOR_ATTACHMENT1_EXT); //Draw into the second texture
		
		glBindTexture(GL_TEXTURE_2D, texID[0]); // read from the first texture
		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, size, size, texFormat, GL_FLOAT, h);
		
		//checkGLErrors ("display 2");

		glUseProgram(shader_prog);

		//checkGLErrors ("display 3");

		glUniform1i(texHandle, 0);
		glUniform1i(sizeHandle, size);
		glUniform1f(delta_sizeHandle, delta_size);
		//glUniform1f(delta_sizeHandle, delta_t);

		//checkGLErrors ("display 4");

		glBegin(GL_QUADS);
            glTexCoord2f(0, 0); glVertex2f(0.0f, 0.0f);
            glTexCoord2f(1, 0); glVertex2f((GLfloat)size, 0.0f);
            glTexCoord2f(1, 1); glVertex2f((GLfloat)size, (GLfloat)size);
            glTexCoord2f(0, 1); glVertex2f(0.0f, (GLfloat)size);
        glEnd();

		glUseProgram(0);


		// display
		
		//checkGLErrors ("display 5");

		
		// Bind the filtered texture
        //glBindTexture(GL_TEXTURE_2D, texID[1]); //read from the second texture

		glReadBuffer(GL_COLOR_ATTACHMENT1_EXT);
		glReadPixels(0, 0, size, size, texFormat,GL_FLOAT, h_temp2);

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

/*		glViewport(0, 0, 200, 200);
		glDrawBuffer(currentDrawbuf);
		glEnable(GL_TEXTURE_2D);

		glBegin(GL_QUADS);
            glTexCoord2f(0, 0); glVertex3f(0.0f, 0.0f, 0.0f);
            glTexCoord2f(1, 0); glVertex3f((GLfloat)size, 0.0f, 0.0f);
            glTexCoord2f(1, 1); glVertex3f((GLfloat)size, (GLfloat)size, 0.0f);
            glTexCoord2f(0, 1); glVertex3f(0.0f, (GLfloat)size, 0.0f);
        glEnd();

        glDisable(GL_TEXTURE_2D);
		checkGLErrors ("END: display");
*/
		//glMatrixMode(GL_PROJECTION);
		//glLoadIdentity();
		//glFrustum (-1.0f, 1.0f, -1.0f, 1.0f, 1.5f, 500.0f);
		//glMatrixMode(GL_MODELVIEW);
		//glLoadIdentity();
		
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glViewport(0, 0, 800, 800);

		glEnable(GL_CULL_FACE);
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_LIGHTING);
	}


	glutSwapBuffers();

	totalDraw += glutGet(GLUT_ELAPSED_TIME) - startDraw;	

	static bool stats_10sec_printed = false;
	static GLint totalTime = 0;
	GLint startTime = glutGet(GLUT_ELAPSED_TIME);

	if (!stats_10sec_printed && totalTime > 10000) {
		printStats();
		stats_10sec_printed = true;
	}

	static GLint oldTime = glutGet(GLUT_ELAPSED_TIME);
	static GLint timebase = oldTime;
	static GLint frame = 0;

	GLint newTime = glutGet(GLUT_ELAPSED_TIME);
	frame++;

	//std::cout << "Elapsed time: " << newTime - oldTime << std::endl;

	if (oldTime == newTime)
		return;

	if (newTime - timebase > 1000) {
		if (printFPS)
			std::cout << "FPS: " << (GLint)(frame*1000.0f/(newTime-timebase)) << std::endl;
		timebase = newTime;
		frame = 0;
	}

	if (!pause) {
		// rain
		if (rain) {
			GLint chance = (GLint) (((GLfloat) rand() / (GLfloat) RAND_MAX) * 10.0f);
			if (chance > 6) {
				GLint i = (GLint) (((GLfloat) rand() / (GLfloat) RAND_MAX) * (GLfloat) (size-2)) + 1;
				GLint j = (GLint) (((GLfloat) rand() / (GLfloat) RAND_MAX) * (GLfloat) (size-2)) + 1;
				//std::cout << "i: " << i << "  j: " << j << std::endl;
				if (i > 0) h[((i-1)*size + j)*4] += .5f; else h[(i*size + j)*4] += .5f;
				if (j > 0) h[(i*size + (j-1))*4] += .5f; else h[(i*size + j)*4] += .5f;
				if (i > 0 && j > 0) h[((i-1)*size + (j-1))*4] += .5f; else h[(i*size + j)*4] += .5f;
				if (j > 0 && i < size-2) h[((i+1)*size + (j-1))*4] += .5f; else h[(i*size + j)*4] += .5f;
				h[(i*size + j)*4] -= 4.0f;
				if (i < size-2) h[((i+1)*size+j)*4] += .5f; else h[(i*size + j)*4] += 1.0f;
				if (j < size-2) h[(i*size+(j+1))*4] += .5f; else h[(i*size + j)*4] += 1.0f;
				if (i < size-2 && j < size-2) h[((i+1)*size+(j+1))*4] += .5f; else h[(i*size + j)*4] += .5f;
				if (j < size-2 && i > 0) h[((i-1)*size+(j+1))*4] += .5f; else h[(i*size + j)*4] += .5f;
			}
		}

		wave_algorithm(newTime - oldTime);
	}
	oldTime = newTime;

	totalTime += glutGet(GLUT_ELAPSED_TIME) - startTime;
}

void reshape (GLint w, GLint h)
{
	glViewport (0, 0, (GLsizei) w, (GLsizei) h);
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	glFrustum (-1.0f, 1.0f, -1.0f, 1.0f, 1.5f, 500.0f);
	glMatrixMode (GL_MODELVIEW);
}

void init2DTexture() {
	checkGLErrors ("START: init2DTexture");
	//float data[size][size][4];
	/*for (GLint i = 0 ; i < size; i++) {
		for (GLint j = 0; j < size; j++) {
			data[i][j][0] = 15.0f;
			data[i][j][1] = 10.0f;
			data[i][j][2] = 8.0f;
			data[i][j][3] = 1.0f;
		}
	}*/
	glGenTextures(2, texID);
	//glActiveTexture(GL_TEXTURE0);
	//checkGLErrors ("init2DTexture  1");

	for (int i = 0; i < 2; i++)	{
		glBindTexture(texTarget, texID[i]);
		// (set texture parameters here)
		glTexParameteri(texTarget, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(texTarget, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(texTarget, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(texTarget, GL_TEXTURE_WRAP_T, GL_CLAMP);
		//create the texture
		glTexImage2D(texTarget, 0, texInternalFormat, size, size, 0, texFormat, GL_FLOAT, 0);
	}

	checkGLErrors ("init2DTexture  2");

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fboId);

	//checkGLErrors ("init2DTexture  3");
	
	//glTexImage2D(texTarget, 0, texInternalFormat, size, size, 0, texFormat, GL_FLOAT, 0);//h);
	//glTexSubImage2D(texTarget, 0, 0, 0, size, size, texFormat, GL_FLOAT, h);

	//glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	//glBindTexture(texTarget, 0);

	glGetIntegerv( GL_FRAMEBUFFER_BINDING_EXT, &savedFboId );
	if (fboId != savedFboId)
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fboId);
	glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, texTarget, texID[0], 0);
	checkGLErrors ("init2DTexture  4");
	if (savedFboId != fboId)
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, (GLuint)savedFboId);

	glGetIntegerv( GL_FRAMEBUFFER_BINDING_EXT, &savedFboId );
	if (fboId != savedFboId)
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fboId);
	checkGLErrors ("init2DTexture  5");
	glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT1_EXT, texTarget, texID[1], 0);
	checkGLErrors ("init2DTexture  6");
	// Returns FBO binding to the previously enabled FBO
	if (savedFboId != fboId)
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, (GLuint)savedFboId);

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	checkGLErrors ("END: init2DTexture");
}

char *textFileRead(char *fn)
{
	FILE *fp;
	char *content = NULL;

	int count=0;

	if (fn != NULL) {
		fp = fopen(fn,"rt");

		if (fp != NULL) {

			fseek(fp, 0, SEEK_END);
			count = ftell(fp);
			rewind(fp);

			if (count > 0) {
				content = (char *)malloc(sizeof(char) * (count+1));
				count = fread(content,sizeof(char),count,fp);
				content[count] = '\0';
			}
			fclose(fp);
		}
	}
	return content;
}

void setShaders()
{
	checkGLErrors ("START: setShaders");
	char *vs, *fs;
	
	v_shader = glCreateShader(GL_VERTEX_SHADER);
	f_shader = glCreateShader(GL_FRAGMENT_SHADER);

	vs = textFileRead("jacobi.vert");
	fs = textFileRead("jacobi.frag");

	const char * vv = vs;
	const char * ff = fs;

	glShaderSource(v_shader, 1, &vv, NULL);
	glShaderSource(f_shader, 1, &ff, NULL);

	free(vs);
	free(fs);

	glCompileShader(v_shader);
	glCompileShader(f_shader);

	shader_prog = glCreateProgram();

	glAttachShader(shader_prog, v_shader);
	glAttachShader(shader_prog, f_shader);

	glLinkProgram(shader_prog);
	glUseProgram(shader_prog);

	texHandle = glGetUniformLocation(shader_prog, "tex");
	sizeHandle = glGetUniformLocation(shader_prog, "size");
	//timeHandle = glGetUniformLocation(shader_prog, "delta_t");
	delta_sizeHandle = glGetUniformLocation(shader_prog, "delta_size");
	//glUniform1i(texHandle, 0);
	//glUseProgram(0);
	checkGLErrors ("END: setShaders");
}

void keyboard (unsigned char key, GLint x_coord, GLint y_coord)
{
	switch (key) {
		/*case 'c':
			cg_on = !cg_on;
			if (cg_on)
				std::cout << "Switched to CG\n";
			else
				std::cout << "Switched to Jacobi\n";
			break;*/
		case 'q':
		case 27:
			exit(0);
		case 'r':
			rain = !rain;
			break;
		case 'z':
			pause = !pause;
			for (GLint i = 0 ; i < size; i++) {
				for (GLint j = 0; j < size; j++) {
					x(i,j) = i*delta_size;
					z(i,j) = j*delta_size;
					h[(i*size+j)*4] = h0;
					h[(i*size+j)*4+1] = u0;
					h[(i*size+j)*4+2] = v0;
					normals[i][j] = vector<GLfloat>(3);
				}
			}
			for (GLint i = 8; i < 15; i++) {
				h[i*size*4+6*4] = 2.0f;
				h[i*size*4+7*4] = 2.0f;
				h[i*size*4+8*4] = 2.0f;
				h[i*size*4+9*4] = 2.0f;
				h[i*size*4+10*4] = 2.0f;
				h[i*size*4+11*4] = 2.0f;
				h[i*size*4+12*4] = 2.0f;
				h[i*size*4+13*4] = 2.0f;
				h[i*size*4+14*4] = 2.0f;
				h[i*size*4+15*4] = 2.0f;
			}

			/*for (GLint i = 20; i < 25; i++) {
				h[i][40][0] = 30.0f;
				h[i][41][0] = 30.0f;
				h[i][42][0] = 30.0f;
				h[i][43][0] = 30.0f;
				h[i][44][0] = 30.0f;
				h[i][45][0] = 30.0f;
				h[i][46][0] = 30.0f;
			}*/
			break;
		case 's':
			if (frontmode == GL_FILL)
				frontmode = GL_LINE;
			else if (frontmode == GL_LINE)
				frontmode = GL_POINT;
			else
				frontmode = GL_FILL;
			break;
		case 'f':
			printFPS = !printFPS;
			break;
		case 'w':
			well = !well;
		break;
		case 'p':
			pause = !pause;
			if (pause) std::cout << "Pause stopped\n";
			else       std::cout << "Paused\n";
			break;
		case 'n':
			showNormals = !showNormals;
			break;
		case 'h':
			std::cout << "Keyboard commands:\n" <<
				"-----------------------------------\n" <<
				"q, <esc> - Quit\n" <<
				"h - Help\n" <<
				"p - Pause\n" <<
				"s - Toggle fill/line/point display mode\n" <<
				"r - Toggle on/off rain\n" <<
				"f - Toggle on/off printing of FPS\n" <<
				"w - Toggle on/off well\n" <<
				"n - Toggle on/off normals\n" <<
				"c - Toggle between CG and Jacobi\n" <<
				"-----------------------------------\n";
			break;
		default:
			break;
	}
}

GLint main(GLint argc, char** argv)
{
	//glutInit(&argc, argv);
	glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	glutInitWindowSize (800, 800);
	glutInitWindowPosition (400, 150);
	glutCreateWindow (argv[0]);

	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutIdleFunc(idle);
	glutKeyboardFunc(keyboard);
	
	init ();
	
	if (useShaders) {
		glewInit();

		if (glewIsSupported("GL_VERSION_2_0"))
			printf("Ready for OpenGL 2.0\n");
		else {
			printf("OpenGL 2.0 not supported\n");
			exit(1);
		}

		glGenFramebuffersEXT(1, &fboId);

		glGetIntegerv( GL_FRAMEBUFFER_BINDING_EXT, &savedFboId );
		if (fboId != savedFboId) {
			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fboId);
		}
		// Returns FBO binding to the previously enabled FBO
		if (savedFboId != fboId) {
			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, (GLuint)savedFboId);
		}

		init2DTexture();
		setShaders();
	}

	glutMainLoop();

	/*if (useShaders) {
		glDeleteTextures (2, texID);
	}*/
	return 0;
}