uniform vec3 EyeCoord;
uniform float WaterSize;
uniform sampler2D DisplacementMap;

const float Eta = 0.75;         // Ratio of indices of refraction 
const float FresnelPower = 2.0;
const float HalfSize = WaterSize / 2.0;

const float F  = 0.02040816; // ((1.0-Eta) * (1.0-Eta)) / ((1.0+Eta) * (1.0+Eta));

varying vec3  Reflect;
varying vec3  Refract;
varying float Ratio;
varying bool  Wall;

void main(void)
{
	vec4 newVertexPos;
	vec4 dv;
	float df;
	
	dv = texture2D( DisplacementMap, gl_MultiTexCoord0.xy );
	
	df = 0.30*dv.x + 0.59*dv.y + 0.11*dv.z;
	
	//newVertexPos = vec4(gl_Normal * df * 100.0, 0.0) + gl_Vertex;

	vec4 ecPosition  = gl_ModelViewMatrix * gl_Vertex;
    vec3 ecPosition3 = ecPosition.xyz / ecPosition.w;
    
    vec3 i = normalize(ecPosition3);
    vec3 n = normalize(gl_NormalMatrix * gl_Normal);
    
    
	Ratio = min(F + (1.0 - F) * pow((1.0 - dot(-i, n)), FresnelPower), 1.0);

    vec3 refr = refract(normalize(gl_Vertex.xyz - EyeCoord), normalize(gl_Normal), Eta);
    vec3 refl = reflect(normalize(gl_Vertex.xyz - EyeCoord), normalize(gl_Normal));
    
    float dist;
    vec3 refl_point, refr_point;
    
    if (refl.x < 0.0) {
    	dist = -gl_Vertex.x / refl.x;
    	refl_point = gl_Vertex.xyz + dist * refl;
    } else {
    	dist = (WaterSize - 1.0 - gl_Vertex.x) / refl.x;
    	refl_point = gl_Vertex.xyz + dist * refl;
    }
    if (refl_point.y < 0.0) {
    	dist = (-HalfSize - gl_Vertex.y) / refl.y;
    	refl_point = gl_Vertex.xyz + dist * refl;
    } else if (refl_point.y > WaterSize - 1.0) {
    	dist = (HalfSize - gl_Vertex.y) / refl.y;
    	refl_point = gl_Vertex.xyz + dist * refl;
    }
    if (refl_point.z < 0.0) {
    	dist = -gl_Vertex.z / refl.z;
    	refl_point = gl_Vertex.xyz + dist * refl;
    } else if (refl_point.z > WaterSize - 1.0) {
    	dist = (WaterSize - 1.0 - gl_Vertex.z) / refl.z;
    	refl_point = gl_Vertex.xyz + dist * refl;
    }
    
    if (refr.x < 0.0) {
    	dist = -gl_Vertex.x / refr.x;
    	refr_point = gl_Vertex.xyz + dist * refr;
    } else {
    	dist = (WaterSize - 1.0 - gl_Vertex.x) / refr.x;
    	refr_point = gl_Vertex.xyz + dist * refr;
    }
    if (refr_point.y < 0.0) {
    	dist = (-HalfSize - gl_Vertex.y) / refr.y;
    	refr_point = gl_Vertex.xyz + dist * refr;
    } else if (refr_point.y > WaterSize - 1.0) {
    	dist = (HalfSize - gl_Vertex.y) / refr.y;
    	refr_point = gl_Vertex.xyz + dist * refr;
    }
    if (refr_point.z < 0.0) {
    	dist = -gl_Vertex.z / refr.z;
    	refr_point = gl_Vertex.xyz + dist * refr;
    } else if (refr_point.z > WaterSize - 1.0) {
    	dist = (WaterSize - 1.0 - gl_Vertex.z) / refr.z;
    	refr_point = gl_Vertex.xyz + dist * refr;
    }

//    if (refl_point.y <= 18.0) {
    	Reflect = (gl_ModelViewMatrix * (vec4(refl_point, 1) - vec4(HalfSize-0.5, 0.0, HalfSize-0.5, 1.0))).xyz;
    	Wall = true;
//    } else {
//    	Reflect = reflect(i, n);
//    	Wall = false;
//    }
    Reflect = vec3(gl_TextureMatrix[0] * vec4(Reflect, 1.0));
    
    Refract = (gl_ModelViewMatrix * (vec4(refr_point, 1) - vec4(HalfSize, 0.0, HalfSize, 1.0))).xyz;
    Refract = vec3(gl_TextureMatrix[0] * vec4(Refract, 1.0));

    gl_Position = ftransform();
}
